/* color.c */
/* Copyright 1999 by Steve Kirkendall */

#include "elvis.h"

typedef struct guidot_s
{
	struct guidot_s *next;
	CHAR	*name;	/* "gui.role" form of some other gui's colors */
	CHAR	*descr;	/* description of the attributes */
} guidot_t;

COLORINFO colorinfo[128] =
{
	{NULL},			/* the null color */
	{toCHAR("normal")},	/* the "normal" color */
	{toCHAR("idle")},	/* the "idle" color */
	{toCHAR("selection")}	/* the "selection" color */
};
static int npermanent = COLOR_FONT_QTY_SPECIALS;
static int ntemporary;	/* index of next temporary colorinfo[] to use */

/* This array is used to store the sort sequence for the colors.  We don't
 * simply store the colorinfo[] array in place because we don't want to move
 * any font after it has been assigned a position in colorinfo[].
 */
static int sortorder[127] = {2, 1, 3};

/* This is the head of a list of colors that would be used in a different
 * GUI.  We store this for the sole purpose of making :mkexrc work better
 * for multiple GUIs.
 */
static guidot_t *guidot;

/* These three are used for storing info about combined colors */
static int prevold;
static int prevnew;
static int prevcombo;

/* Forward declarations for static functions */
static void colorlike P_((int fontcode));
static void recolor P_((int fontcode));

/* Returns the index of a font name.  If the fontname isn't in the colorinfo[]
 * table already, then this adds it.  If it can't add it (because the table is
 * already full) then it outputs an error message and returns 0.
 */
int colorfind(fontname)
	CHAR *fontname;
{
	int	i;

	/* scan the list */
	for (i = 1; i < npermanent; i++)
		if (!CHARcmp(colorinfo[i].name, fontname))
			return i;

	/* must be a new color -- is there room? */
	if (i >= QTY(colorinfo))
	{
		msg(MSG_ERROR, "[S]color table full -- can't add $1", fontname);
		return 0;
	}

	/* allocate the color */
	colorinfo[i] = colorinfo[0];
	colorinfo[i].name = CHARkdup(fontname);
	while (--i >= 1 && CHARcmp(colorinfo[sortorder[i - 1]].name, colorinfo[npermanent].name) > 0)
		sortorder[i] = sortorder[i - 1];
	sortorder[i] = npermanent;
	return npermanent++;
}

/* Combine two fonts, and return the result in a static variable */
COLORINFO *colorcombine(oldfont, newcinfo)
	int	oldfont;	/* index of the old font */
	COLORINFO *newcinfo;	/* information about new font */
{
	static COLORINFO combo;

	if (&combo != newcinfo)
		combo = *newcinfo;
	if (~combo.bits & COLOR_FG)
		combo.fg = colorinfo[oldfont].fg;
	if (~combo.bits & COLOR_BG)
		combo.bg = colorinfo[oldfont].bg;
	if (combo.bits & COLOR_PROPSET)
		combo.bits |= (colorinfo[oldfont].bits & ~(COLOR_SET|COLOR_PROP));
	else
		combo.bits |= (colorinfo[oldfont].bits & ~(COLOR_SET|COLOR_PROPSET));
	return &combo;
}

/* Interpret the "descr" string for the given fontcode.  Then call
 * colorlike(fontcode) to adjust any fonts which are "like" this one.
 */
static void recolor(fontcode)
	int	fontcode;
{
	CHAR	*scan, *build, *word;
	COLORINFO newinfo, *combo;
	enum	{EXPECT_LIKE, EXPECT_FG, EXPECT_BG} expect, nextexpect;
	long	oldfg, oldbg;
	int	oldbits, i;
	BOOLEAN	using;

	/* reset the newinfo */
	newinfo = colorinfo[0];

	/* if there is a description for this font, then parse it */
	if (colorinfo[fontcode].descr)
	{
		build = word = NULL;
		expect = nextexpect = EXPECT_FG;
		for (scan = colorinfo[fontcode].descr; ; scan++)
		{
			/* part of a word? */
			if (*scan && !isspace(*scan))
			{
				(void)buildCHAR(&build, *scan);
				if (!word)
					word = build;
				goto Continue;
			}	

			/* Whitespace. If after other whitespace, then ignore */
			if (!build)
				goto Continue;

			/* okay, we've hit the end of a word.  If we were
			 * expecting a role name, then this is it.
			 */
			if (expect == EXPECT_LIKE)
			{
				newinfo.like = colorfind(build);
				expect = EXPECT_FG;
				build = word = NULL;
				goto Continue;
			}

			/* Is it a word that has special meaning? */
			if (!CHARcmp(word, toCHAR("like")))
				nextexpect = EXPECT_LIKE;
			else if (!CHARcmp(word, toCHAR("on")))
				nextexpect = EXPECT_BG;
			else if (!CHARcmp(word, toCHAR("bold")))
				newinfo.bits |= COLOR_BOLD;
			else if (!CHARcmp(word, toCHAR("italic")))
				newinfo.bits |= COLOR_ITALIC;
			else if (!CHARcmp(word, toCHAR("underlined")))
				newinfo.bits |= COLOR_UNDERLINED;
			else if (!CHARcmp(word, toCHAR("boxed")))
				newinfo.bits |= COLOR_BOXED;
			else if (!CHARcmp(word, toCHAR("proportional")))
				newinfo.bits |= COLOR_PROP|COLOR_PROPSET;
			else if (!CHARcmp(word, toCHAR("fixed")))
				newinfo.bits = (newinfo.bits & ~COLOR_PROP)
							| COLOR_PROPSET;
			else if (!CHARcmp(word, toCHAR("graphic")))
				newinfo.bits |= COLOR_GRAPHIC;
			else
				word = NULL;

			/* If the preceeding word was special or we've hit the
			 * end of the descr string, and some other text
			 * preceeded the word, then that other text must be
			 * some kind of argument.
			 */
			if ((word || !*scan) && build != word)
			{
				/* If a special word appeared, strip it from
				 * the end of the other text.
				 */
				if (word)
					word[-1] = '\0';

				/* Strip any trailing whitespace.  (This can
				 * happen at the end of the command string.)
				 */
				for (i = CHARlen(build); isspace(build[--i]); )
				{
				}
				build[i + 1] = '\0';

				/* What kind of text is it? */
				switch (expect)
				{
				  case EXPECT_LIKE:
					/* can't happen -- handled above */
					break;

				  case EXPECT_FG:
					/* After nothing in particular, we
					 * expect a foreground color.  Convert
					 * it to a number; if successful, then
					 * set newinfo's COLOR_FG bit.
					 */
					if (gui->color
					 && (*gui->color)(fontcode, build, True,
						  &newinfo.fg, newinfo.fg_rgb))
						newinfo.bits |= COLOR_FG;
					break;

				  case EXPECT_BG:
					/* After "on", we expect a background
					 * color.  Convert it to a number; if
					 * successful, then set newinfo's
					 * COLOR_BG bit.
					 */
					if (gui->color
					 && (*gui->color)(fontcode, build,False,
						  &newinfo.bg, newinfo.bg_rgb))
						newinfo.bits |= COLOR_BG;
					break;
				}
				expect = nextexpect;
			}

			/* If we're continuing a long arg such as a color name,
			 * then add a space.  Otherwise free the string.
			 */
			if (!word && *scan)
			{
				/* continuing, so add a space */
				buildCHAR(&build, ' ');
				word = build + CHARlen(build);
			}
			else
			{
				/* free the old string, prepare for a new one */
				safefree(build);
				build = word = NULL;
				expect = nextexpect;

Continue:
				/* if end of string, then quit */
				if (!*scan)
					break;
			}

		} /* end for each char of description */
	} /* end if there is a description */

	/* Remember the old colors, so we can free them later */
	oldfg = colorinfo[fontcode].fg;
	oldbg = colorinfo[fontcode].bg;
	oldbits = colorinfo[fontcode].bits;

	/* Can't force off the colors for "normal" */
	if (fontcode == COLOR_FONT_NORMAL)
	{
		if (~newinfo.bits & COLOR_FG)
		{
			newinfo.fg = oldfg;
			memcpy(newinfo.fg_rgb, colorinfo[fontcode].fg_rgb, 3);
			newinfo.bits |= COLOR_FG;
		}
		if (~newinfo.bits & COLOR_BG)
		{
			newinfo.bg = oldbg;
			memcpy(newinfo.bg_rgb, colorinfo[fontcode].bg_rgb, 3);
			newinfo.bits |= COLOR_BG;
		}
	}

	/* combine the new info with the "like" info, and store it */
	combo = colorcombine(newinfo.like, &newinfo);
	colorinfo[fontcode].fg = combo->fg;
	colorinfo[fontcode].bg = combo->bg;
	colorinfo[fontcode].bits = combo->bits;
	memcpy(colorinfo[fontcode].fg_rgb, newinfo.fg_rgb, 3);
	memcpy(colorinfo[fontcode].bg_rgb, newinfo.bg_rgb, 3);

	/* we also store the "like" value for later updates, unless it is
	 * like itself or one of the built-ins.  Exception: "idle" can be
	 * "like normal".
	 */
	colorinfo[fontcode].like = newinfo.like;
	if (newinfo.like == COLOR_FONT_IDLE
	 || (newinfo.like == COLOR_FONT_NORMAL && fontcode != COLOR_FONT_IDLE)
	 || newinfo.like == fontcode)
		colorinfo[fontcode].like = 0;

	/* The "normal" colors are inherited by the other special color codes,
	 * unless they've been explicitly set elsewhere.  Inheriting them
	 * this way causes them to be considered "set", so really this only
	 * happens the first time you set "normal" colors, or every time you
	 * set colors for the other special fonts.
	 */
	if (fontcode == COLOR_FONT_NORMAL || fontcode == COLOR_FONT_IDLE)
	{
		for (i = COLOR_FONT_NORMAL + 1; i < COLOR_FONT_QTY_SPECIALS; i++)
		{
			if (~colorinfo[i].bits & COLOR_FG)
				colorinfo[i].fg = colorinfo[COLOR_FONT_NORMAL].fg;
			if (~colorinfo[i].bits & COLOR_BG)
				colorinfo[i].bg = colorinfo[COLOR_FONT_NORMAL].bg;
			colorinfo[i].bits |= (colorinfo[COLOR_FONT_NORMAL].bits & (COLOR_FG|COLOR_BG));
		}
	}
	else if (fontcode < COLOR_FONT_QTY_SPECIALS)
	{
		if (~colorinfo[fontcode].bits & COLOR_FG)
			colorinfo[fontcode].fg = colorinfo[COLOR_FONT_NORMAL].fg;
		if (~colorinfo[fontcode].bits & COLOR_BG)
			colorinfo[fontcode].bg = colorinfo[COLOR_FONT_NORMAL].bg;
		colorinfo[fontcode].bits |= (colorinfo[COLOR_FONT_NORMAL].bits & (COLOR_FG|COLOR_BG));
	}

	/* if this GUI has a freecolor() function, then decide whether to
	 * free any colors.
	 */
	if (gui->freecolor)
	{
		/* maybe free the old foreground color */
		if (oldbits & COLOR_FG)
		{
			for (i=1, using=False; i < npermanent && !using; i++)
				if (colorinfo[i].fg == oldfg
				 || colorinfo[i].bg == oldfg)
					using = True;
			if (!using)
			{
				(*gui->freecolor)(oldfg, True);
			}
		}

		/* maybe free the old background color */
		if (oldbits & COLOR_BG)
		{
			for (i=1, using=False; i < npermanent && !using; i++)
				if (colorinfo[i].fg == oldbg
				 || colorinfo[i].bg == oldbg)
					using = True;
			if (!using)
			{
				(*gui->freecolor)(oldbg, False);
			}
		}
	}

	/* call colorlike() to adjust any colors which are "like" this one */
	colorlike(fontcode);
}

/* Find each entry in the color table which has "like" that matches "fontcode",
 * and reinterpret its "descr" field by calling recolor() on it.
 */
static void colorlike(fontcode)
	int	fontcode;
{
	int	i;
	int	oldlike;

	/* Set this font's "like" value to an impossible value temporarily,
	 * to avoid any chance of an endless loop due to cyclic dependencies.
	 */
	oldlike = colorinfo[fontcode].like;
	colorinfo[fontcode].like = 0;

	/* adjust any other colors which are "like" this one */
	for (i = 1; i < npermanent; i++)
		if (colorinfo[i].like == fontcode)
			recolor(i);

	/* restore this font's "like" value */
	colorinfo[fontcode].like = oldlike;
}


/* Change the description for a given fontcode, and then recolor it so the
 * changes take effect.  The "explicit" parameter should be True if called
 * from :color, or False if set to a default value via elvis.syn or something
 * similar.
 */
void colorset(fontcode, descr, explicit)
	int	fontcode;
	CHAR	*descr;
	BOOLEAN	explicit;
{
	BOOLEAN	washidden;

	/* If not explicit, and a previous colorset was explicit, then do
	 * nothing here.  Defaults never override explicit settings.
	 */
	if (!explicit && (colorinfo[fontcode].bits & COLOR_SET) != 0)
		return;

	/* If there is an old description, free it now.  Use the new one. */
	if (colorinfo[fontcode].descr)
		safefree(colorinfo[fontcode].descr);
	colorinfo[fontcode].descr = CHARkdup(descr);

	/* Call recolor() to do the hard work.  If this is a non-explicit
	 * setting, then disable error messages while it runs, since there's
	 * no good reason to badger the user for my mistakes.
	 */
	if (explicit)
		recolor(fontcode);
	else
	{
		washidden = msghide(True);
		recolor(fontcode);
		(void)msghide(washidden);
	}

	/* force the COLOR_SET bit on, if necessary */
	if (explicit)
		colorinfo[fontcode].bits |= COLOR_SET;
}

/* Set a color for a GUI other than the one we're using.  This has no effect
 * except that it is incorporated into the :mkexrc script.
 */
void colorforeign(name, descr)
	CHAR	*name;	/* name, formatted as "gui.role" */
	CHAR	*descr;	/* the attributes for this color role */
{
	guidot_t *scan, *newp;

	/* If already set, then change the description */
	for (scan = guidot; scan; scan = scan->next)
	{
		if (!CHARcmp(scan->name, name))
		{
			safefree(scan->descr);
			scan->descr = CHARdup(descr);
			return;
		}
	}

	/* else insert a new entry at the tail of the list */
	newp = safealloc(1, sizeof(guidot_t));
	newp->name = CHARdup(name);
	newp->descr = CHARdup(descr);
	newp->next = NULL;
	if (!guidot)
		guidot = newp;
	else
	{
		for (scan = guidot; scan->next; scan = scan->next)
		{
		}
		newp->next = scan->next;
		scan->next = newp;
	}
}


/* Reset the temporary color combination counter.  This is called by the
 * drawimage() function before the display->setup() function.
 */
void colorsetup P_((void))
{
	/* Set ntemporary to start allocating immediately after the permanent
	 * fonts.  Also set prevnew to an impossible value so the prevXXX
	 * variables won't come into play.
	 */
	ntemporary = npermanent;
	prevnew = 0;
}

/* If possible, allocate a new temporary font for storing the result of
 * combining oldfont and newfont, and return the combined font's index.  If
 * that isn't possible (because we ran out of space in the color table) then
 * just return the newfont.
 */
int colortmp(oldfont, newfont)
	int	oldfont;
	int	newfont;
{
	/* if same as previous, then reuse it */
	if (oldfont == prevold && newfont == prevnew)
		return prevcombo;

	/* if no more room for combinations, then return the newfont.  Also use
	 * newfont if there is no oldfont, or we're combining a font with
	 * itself.
	 */
	if (oldfont == 0 || oldfont == newfont || ntemporary >= QTY(colorinfo))
		return newfont;

	/* combine the fonts to form a new one */
	colorinfo[ntemporary] = *colorcombine(oldfont, &colorinfo[newfont]);

	/* remember info to help optimimize the next call */
	prevold = oldfont;
	prevnew = newfont;
	prevcombo = ntemporary++;

	/* return the combined font */
	return prevcombo;
}

/* List the current color definitions out to a given window */
void colorlist(win, name)
	WINDOW	win;	/* which window to write to */
	CHAR	*name;	/* name of color, or NULL for all */
{
	int	sorted, i;
	BOOLEAN	found;

	for (sorted = 0, found = False; sorted < npermanent - 1; sorted++)
	{
		/* skip if not the named color */
		i = sortorder[sorted];
		if (name ? CHARcmp(name, colorinfo[i].name)
			 : (~colorinfo[i].bits & COLOR_SET))
			continue;

		/* output the description */
		if (colorinfo[i].descr)
			msg(MSG_INFO, "[SS]color $1 $2",
				colorinfo[i].name, colorinfo[i].descr);
		else
			msg(MSG_INFO, "[S]color $1 has not been set",
				colorinfo[i].name);
		found = True;
	}
	if (name && !found)
		msg(MSG_INFO, "[S]nothing is using colors for $1", name);
}

#ifdef FEATURE_MKEXRC
/* Append a series of ex commands to "buf", to recreate the current color
 * scheme.  Ignore any colors which don't have the COLOR_SET bit set.
 */
void colorsave(buf)
	BUFFER	buf;
{
	int	i;
	CHAR	cmd[200];
	MARKBUF	end;
	guidot_t *scan;

	/* for each color role... */
	for (i = 1; i < npermanent; i++)
	{
		/* skip if only set to a default value */
		if ((colorinfo[i].bits & COLOR_SET) == 0)
			continue;

		/* construct a command line */
		CHARcpy(cmd, toCHAR("color "));
		CHARcat(cmd, o_gui);
		CHARcat(cmd, toCHAR("."));
		CHARcat(cmd, colorinfo[i].name);
		CHARcat(cmd, toCHAR(" "));
		CHARcat(cmd, colorinfo[i].descr);
		CHARcat(cmd, toCHAR("\n"));

		/* append it to the buffer */
		(void)marktmp(end, buf, o_bufchars(buf));
		bufreplace(&end, &end, cmd, CHARlen(cmd));
	}

	/* Append the colors for other GUIs too */
	for (scan = guidot; scan; scan = scan->next)
	{
		CHARcpy(cmd, toCHAR("color "));
		CHARcat(cmd, scan->name);
		CHARcat(cmd, toCHAR(" "));
		CHARcat(cmd, scan->descr);
		CHARcat(cmd, toCHAR("\n"));

		/* append it to the buffer */
		(void)marktmp(end, buf, o_bufchars(buf));
		bufreplace(&end, &end, cmd, CHARlen(cmd));
	}
}
#endif /* FEATURE_MKEXRC */

#ifdef FEATURE_COMPLETE
/* Attempt to complete an argument to :color.  This may be role name, or one
 * of the special words.  The names of actual colors are not expanded.  As a
 * last resort, it may return the current args.
 */
CHAR *colorcomplete(win, from, to)
	WINDOW	win;	/* where to list the possible matches */
	MARK	from;	/* start of args */
	MARK	to;	/* end of args, where completion takes place */
{
	CHAR	*cp, *word, *role;
	long	len, width;
	BOOLEAN	wantrole;
	int	i, sorted;
	int	matches, firstmatch, common;
 static CHAR	retbuf[100];
 static CHAR	*magic[] = {
		toCHAR("bold "),
		toCHAR("boxed "),
		toCHAR("proportional "),
		toCHAR("fixed "),
		toCHAR("graphic "),
		toCHAR("italic "),
		toCHAR("like "),
		toCHAR("on "),
		toCHAR("underlined ")};
	

	/* scan the words leading up to current word */
	word = NULL;
	wantrole = True;
	role = NULL;
	len = markoffset(to) - markoffset(from);
	for (scanalloc(&cp, from); cp && len > 0; scannext(&cp), len--)
	{
		/* if in word, accumulate chars */
		if (!isspace(*cp))
		{
			buildCHAR(&word, *cp);
			continue;
		}

		/* ignore redundant spaces */
		if (!word)
			continue;

		/* end of a complete word -- if it was "like" then we expect
		 * a role. */
		wantrole = (BOOLEAN)!CHARcmp(word, toCHAR("like"));
		if (!role)
			role = word;
		else
			safefree(word);
		word = NULL;
	}
	scanfree(&cp);

	/* At this point, we've reached the point where completion is to take
	 * place.  If we're in the middle of a partial word, then "word" points
	 * to that partial word.  The "wantrole" variable indicates whether
	 * we'll be completing a role name or something else.
	 */

	/* Are we completing a role name? */
	if (wantrole)
	{
		/* we don't care any complete role name */
		if (role)
			safefree(role);

		/* if no chars, then list all roles */
		if (!word)
		{
			width = o_columns(win);
			for (sorted = 0; sorted < npermanent - 1; sorted++)
			{
				i = sortorder[sorted];
				if (CHARlen(colorinfo[i].name) + 2 > width)
				{
					drawextext(win, toCHAR("\n"), 1);
					width = o_columns(win);
				}
				else if (width < o_columns(win))
				{
					drawextext(win, toCHAR(" "), 1);
					width--;
				}
				drawextext(win, colorinfo[i].name, CHARlen(colorinfo[i].name));
				width -= CHARlen(colorinfo[i].name);
			}
			drawextext(win, toCHAR("\n"), 1);
			return toCHAR("");
		}

		/* see how many matches there are, and how many more characters
		 * they have in common.
		 */
		matches = 0;
		len = CHARlen(word);
		for (i = 1; i < npermanent; i++)
		{
			/* skip non-matching role names */
			if (CHARncmp(word, colorinfo[i].name, len))
				continue;

			/* first match? */
			if (matches++ == 0)
			{
				/* yes, remember it */
				firstmatch = i;
				common = CHARlen(colorinfo[i].name);
			}
			else
			{
				/* no, check common chars */
				while (CHARncmp(colorinfo[firstmatch].name,
					 	colorinfo[i].name, common))
					common--;
			}
		}

		/* don't need "word" anymore */
		safefree(word);

		/* If no matches, then beep and return nothing. */
		if (matches == 0)
		{
			guibeep(win);
			return toCHAR("");
		}

		/* If unique match, then return its tail plus a space */
		if (matches == 1)
		{
			CHARcpy(retbuf, colorinfo[firstmatch].name + len);
			CHARcat(retbuf, toCHAR(" "));
			return retbuf;
		}

		/* It isn't unique, but maybe we can complete some chars */
		if (common > len)
		{
			CHARcpy(retbuf, colorinfo[firstmatch].name + len);
			retbuf[common - len] = '\0';
			return retbuf;
		}

		/* List the partially-matching role names */
		width = o_columns(win);
		for (sorted = 0; sorted < npermanent - 1; sorted++)
		{
			/* skip non-matching names */
			i = sortorder[sorted];
			if (CHARncmp(colorinfo[i].name, colorinfo[firstmatch].name, len))
				continue;
			if (CHARlen(colorinfo[i].name) + 2 > width)
			{
				drawextext(win, toCHAR("\n"), 1);
				width = o_columns(win);
			}
			else if (width < o_columns(win))
			{
				drawextext(win, toCHAR(" "), 1);
				width--;
			}
			drawextext(win, colorinfo[i].name, CHARlen(colorinfo[i].name));
			width -= CHARlen(colorinfo[i].name);
		}
		drawextext(win, toCHAR("\n"), 1);
		return toCHAR("");
	}

	/* if we get here, then it wasn't a role name, so we want to complete
	 * other stuff.
	 */

	/* if no partial word, then we return this role's settings.  Note that
	 * we are guaranteed to know the role name, since otherwise we would
	 * have performed role name completion, above.
	 */
	if (!word)
	{
		/* locate this role */
		for (i = 1; i < npermanent; i++)
			if (!CHARcmp(role, colorinfo[i].name))
				break;
		safefree(role);
		
		/* if not found, or has no settings, then just beep */
		if (i >= npermanent || !colorinfo[i].descr)
		{
			guibeep(win);
			return toCHAR("");
		}

		/* else return the settings */
		return colorinfo[i].descr;
	}

	/* don't care about role drom here on out */
	safefree(role);

	/* if we get here, then we have a partial word.  Since we can't complete
	 * color names, we can only hope to complete magic words.  All of them
	 * start with different letters except "bold" and "boxed", so this is
	 * easy.
	 */
	if (!CHARcmp(word, toCHAR("b")))
		return toCHAR("o");
	else if (!CHARcmp(word, toCHAR("bo")))
	{
		drawextext(win, toCHAR("bold boxed\n"), 11);
		return toCHAR("");
	}
	len = CHARlen(word);
	for (i = 0; i < QTY(magic); i++)
		if (!CHARncmp(word, magic[i], len))
			return magic[i] + len;

	/* it didn't match anything. */
	guibeep(win);
	return toCHAR("");
}
#endif /* FEATURE_COMPLETE */
