/*
** gwprint.c -- printing support for MS-Windows elvis.
**
** Copyright 1996, Serge Pirotte
*/

#define CHAR    Char
#define BOOLEAN Boolean
#include "elvis.h"
#undef CHAR
#undef BOOLEAN

#if defined (GUI_WIN32)

#include <windows.h>
#include <commctrl.h>
#include "winelvis.h"
#include "elvisres.h"
#ifndef MB_ICONERROR
# define MB_ICONERROR 0
#endif

PRINTDLG        gwpdlg;
int             gw_printing_ok = 1;

GUI_WINDOW      *pgwp;
    
static char     *szDevice = NULL;
static char     *szDriver = NULL;
static char     *szOutput = NULL;

/* This is an array of fonts with every combination of bold/italic/underline
 * attributes.  They are prescaled for the page format.
 */
static HFONT	pfonts[8];

/* This is the current font, from pfonts[8] */
static HFONT    pCurFont = NULL;
static int      pNewPage = 0;

static HDC      printerDC;
static int      oktoprint = 0;

/* This following are sizes & position, in pixels */
static int      pxOffset, pyOffset;	/* position of upper-left corner */
static int      pxSize, pySize;		/* page size */
static int      pxCSize, pyCSize;	/* character size */
static int      pCurX, pCurY;		/* current cursor location */

/* --------------------------------------------------------------------
**
** gw_disable_printing  --  disable printing items.
*/

void gw_disable_printing (GUI_WINDOW *gwp)

{
    HMENU       hMenu;

    hMenu = GetMenu (gwp->frameHWnd);
    EnableMenuItem (hMenu, IDM_FILE_PRINT, MF_GRAYED);
    EnableMenuItem (hMenu, IDM_FILE_PRINTSETUP, MF_GRAYED);
    if (gwp->toolbarHWnd)
        SendMessage (gwp->toolbarHWnd, TB_ENABLEBUTTON, IDM_FILE_PRINT,
                     MAKELONG (FALSE, 0));
}

/* --------------------------------------------------------------------
**
** gw_set_default_printer  --  set the default printer.
*/

void gw_set_default_printer (char *pDevice, char *pDriver, char *pOutput)

{
    if (szDevice != NULL) {
        free (szDevice);
        szDevice = NULL;
	}
	if (pDevice != NULL) {
        szDevice = malloc (strlen (pDevice) + 1);
        strcpy (szDevice, pDevice);
	}
    
    if (szDriver != NULL) {
        free (szDriver);
        szDriver = NULL;
	}
	if (pDriver != NULL) {
        szDriver = malloc (strlen (pDriver) + 1);
        strcpy (szDriver, pDriver);
	}

    if (szOutput != NULL) {
        free (szOutput);
        szOutput = NULL;
	}
	if (pOutput != NULL) {
        szOutput = malloc (strlen (pOutput) + 1);
        strcpy (szOutput, pOutput);
	}
}

/* --------------------------------------------------------------------
**
** gw_get_default_printer  --  get the default printer.
*/

void gw_get_default_printer (void)

{
    DEVNAMES        *dvnp;

    /* fill the structure */
    memset (&gwpdlg, 0, sizeof (PRINTDLG));
    gwpdlg.lStructSize = sizeof (PRINTDLG);
    gwpdlg.Flags = PD_RETURNDEFAULT;

    /* call the common dialog manager */
    if (PrintDlg (&gwpdlg)) {

        /* get device names */
        dvnp = GlobalLock (gwpdlg.hDevNames);
        gw_set_default_printer ((char *)dvnp + dvnp->wDeviceOffset,
                                (char *)dvnp + dvnp->wDriverOffset,
                                (char *)dvnp + dvnp->wOutputOffset);
        GlobalUnlock (gwpdlg.hDevNames);
    }
    else {

        /* disable printing */
        gw_printing_ok = 0;
    }
}

/* --------------------------------------------------------------------
**
** gwbefore  --  prepare a print job.
*/

static void gwbefore (int minor, void (*draw) (_CHAR_ ch))

{
    LOGFONT         lf;
    DOCINFO         di;
    WINDOW          pwin;
    TEXTMETRIC      tm;
    char            optstring[80];
    HDC             dc;
    CHOOSEFONT      cf;
    HFONT           basefont;
    int             i;

    /* check that we have an active Windows window for elvis */
    oktoprint = 0;
    if ((pgwp = gw_find_client (GetFocus ())) == NULL)
        return;

    /* get the associated elvis window, so we can check its options */
    pwin = winofgw ((GUIWIN *)pgwp);

    /* try to create a display context for the printer */
    if ((printerDC = CreateDC (szDriver, szDevice, szOutput, NULL)) == NULL)
        return;
	
    /*  make the display context use transparent background. */
    SetMapMode (printerDC, MM_TEXT);
    SetBkMode (printerDC, TRANSPARENT);

    /* get font type */
    memset (&lf, 0, sizeof (LOGFONT));
    opt_parse_font (o_font (pgwp), &lf);
    lf.lfWeight = FW_NORMAL;
    lf.lfItalic = FALSE;
    lf.lfUnderline = FALSE;
   
    /* Get page size.  Adjust the size to allow the desired number of rows
	 * and columns to fit on the page.
	 */
    pySize = GetDeviceCaps (printerDC, VERTRES);
    pxSize = GetDeviceCaps (printerDC, HORZRES);
    lf.lfHeight = pySize / o_lplines;
    lf.lfWidth = pxSize / o_lpcolumns;
#if 0
	pyOffset = GetDeviceCaps (printerDC, LOGPIXELSY) / 2;
	pxOffset = GetDeviceCaps (printerDC, LOGPIXELSX) / 2;
#else
	pyOffset = 0;
	pxOffset = 0;
#endif

    /* create the fonts */
    basefont = CreateFontIndirect (&lf);
    if (SelectObject (printerDC, basefont) == NULL)
	{
        /* no corresponding user font, ask user to choose */
		DeleteObject (basefont);
        memset (&cf, 0, sizeof (CHOOSEFONT));
        cf.lStructSize = sizeof (CHOOSEFONT);
        cf.hDC = printerDC;
        cf.lpLogFont = &lf;
        cf.Flags = CF_FORCEFONTEXIST | CF_INITTOLOGFONTSTRUCT | CF_PRINTERFONTS;
        ChooseFont (&cf);
        basefont = CreateFontIndirect (&lf);
        if (SelectObject (printerDC, basefont) == NULL)
		{
            ReleaseDC (NULL, printerDC);
            MessageBox (NULL, "Invalid Font For Printer", "WinElvis",
                        MB_ICONERROR | MB_OK);
            return;
        }
    }
	DeleteObject (basefont);

	/* Remember the size of the character cell. */
    GetTextMetrics (printerDC, &tm);
    pxCSize = tm.tmAveCharWidth;
    pyCSize = tm.tmHeight + tm.tmExternalLeading;
 
	/* Generate all varieties of the font with bold/italic/underline */
    for (i = 0; i < QTY(pfonts); i++)
    {
		lf.lfWeight = (i & 1) ? FW_BOLD : FW_NORMAL;
		lf.lfItalic = (i & 2) ? TRUE : FALSE;
		lf.lfUnderline = (i & 4) ? TRUE : FALSE;
		pfonts[i] = CreateFontIndirect (&lf);
    }

    /* The first output will begin a new page. */
    pNewPage = 1;
    pCurX = pxOffset;
    pCurY = pyOffset;
    
    /* inform printer of new job */
    memset (&di, 0, sizeof (DOCINFO));
    di.cbSize = sizeof (DOCINFO);
    di.lpszDocName = (char *)o_filename (pwin->cursor->buffer),
    di.lpszOutput = NULL;
    StartDoc (printerDC, &di);
    
    oktoprint = 1;
}

/* --------------------------------------------------------------------
**
** gwfontch  --  print a char.
*/

static void gwfontch (_char_ fontcode, _CHAR_ ch)

{
    HFONT       pFont;
    COLORREF    fgc;
    int         i;

    /* check if we can print */
    if (!oktoprint)
        return;

    /* check for a new page */
    if (pNewPage)
	{
		StartPage (printerDC);
		pNewPage = 0;
		pCurX = pxOffset;
		pCurY = pyOffset;
    }

    /* select font */
    i = 0;
    if (colorinfo[fontcode].bits & COLOR_BOLD) i += 1;
    if (colorinfo[fontcode].bits & COLOR_ITALIC) i += 2;
    if (colorinfo[fontcode].bits & COLOR_UNDERLINED) i += 4;
    pFont = pfonts[i];

    /* select fg color */
    fgc = o_lpcolor ? colorinfo[fontcode].fg : RGB(0,0,0);

    /* Change the font & color, if necessary */
    if (pFont != pCurFont) {
		pCurFont = pFont;
		SelectObject (printerDC, pCurFont);
		SetTextColor (printerDC, fgc);
    }

    /* print the char */
    if (ch == '\n')
	{
		pCurX = pxOffset;
		pCurY += pyCSize;
    }
    else
	{
		if ((colorinfo[fontcode].bits & COLOR_GRAPHIC) == COLOR_GRAPHIC)
		{
			if (CHARchr("|123456", ch))
			{
				MoveToEx(printerDC, pCurX + pxCSize/2, pCurY + pyCSize/2, NULL);
				LineTo(printerDC, pCurX + pxCSize/2, pCurY);
			}
			if (CHARchr("|456789", ch))
			{
				MoveToEx(printerDC, pCurX + pxCSize/2, pCurY + pyCSize/2, NULL);
				LineTo(printerDC, pCurX + pxCSize/2, pCurY + pyCSize);
			}
			if (CHARchr("-235689", ch))
			{
				MoveToEx(printerDC, pCurX + pxCSize/2, pCurY + pyCSize/2, NULL);
				LineTo(printerDC, pCurX, pCurY + pyCSize/2);
			}
			if (CHARchr("-124578", ch))
			{
				MoveToEx(printerDC, pCurX + pxCSize/2, pCurY + pyCSize/2, NULL);
				LineTo(printerDC, pCurX + pxCSize, pCurY + pyCSize/2);
			}
			if (ch == 'o')
			{
				int radius = pxCSize / 3;
				int xcenter = pCurX + pxCSize / 2;
				int ycenter = pCurY + pyCSize / 2;

				Arc(printerDC, xcenter - radius, ycenter - radius,
					xcenter + radius, ycenter + radius,
					xcenter - radius, ycenter, xcenter - radius, ycenter);
			}
			if (ch == '*')
			{
				int radius = pxCSize / 3;
				int xcenter = pCurX + pxCSize / 2;
				int ycenter = pCurY + pyCSize / 2;
				HBRUSH newbrush, oldbrush;

				newbrush = CreateSolidBrush(fgc);
				oldbrush = SelectObject(printerDC, newbrush);
				Pie(printerDC, xcenter - radius, ycenter - radius,
					xcenter + radius, ycenter + radius,
					xcenter - radius, ycenter, xcenter - radius, ycenter);
				SelectObject(printerDC, oldbrush);
				DeleteObject(newbrush);
			}
		}
		else
		{
			TextOut (printerDC, pCurX, pCurY, (char *)&ch, 1);
		}
		pCurX += pxCSize;
    }
}

/* --------------------------------------------------------------------
**
** gwpage  --  start a new page.
*/

static void gwpage (int linesleft)

{
    if (oktoprint)
	{
	    EndPage (printerDC);
	    pNewPage = 1;
	}
}

/* --------------------------------------------------------------------
**
** gwafter  --  terminate print job.
*/

static void gwafter (int linesleft)

{
    int  i;

    if (oktoprint) {

        /* tell printer it's over */
		EndPage (printerDC);
		EndDoc (printerDC);
	
		/* misc flags */
		pNewPage = 0;
	
		/* delete the fonts. */
		pCurFont = 0;
		for (i = 0; i < QTY(pfonts); i++)
		    DeleteObject (pfonts[i]);
    
		/* delete the device context. */
		DeleteDC (printerDC);
		printerDC = NULL;
		oktoprint = 0;
	}
}

/* -------------------------------------------------------------------- */

LPTYPE  lpwindows = {
    "windows",
    0,
    False,
    gwbefore,
    gwfontch,
    gwpage,
    gwafter
};

#endif
/* ex:set ts=4 sw=4: */
